(* diald monitor TCP client   *)

MODULE DialdMc;
(*$XL+       Modula-2 language extensions: '_' allowed for symbol names *)

FROM   SYSTEM        IMPORT  ADDRESS, ExitProgram, ADR;
FROM   DiagStuff     IMPORT WriteString, WriteLn, RegBreak, DeRegBreak,
                            WrNum, WriteCh, AbEnd,
                            OutHandle, RegExcep, DeRegExcep;
FROM   Conversions   IMPORT StrToCard;
FROM   OS2ARG  IMPORT ArgCount, Arg, STRING, PSTRING;
FROM   OS2DEF  IMPORT APIRET;
FROM   SO32DLL IMPORT EightByte, Socket, AF_INET,
       SOCK_STREAM, INADDR_ANY, Protocol, SockAddr, soclose,
       sock_init, socket, send, connect, recv, psock_errno;
FROM   TCP32DLL IMPORT bswap, gethostbyname, hostent, hostentP;
FROM   DOSEXCEPTIONS IMPORT EXCEPTIONREGISTRATIONRECORD, XCPT_SIGNAL_KILLPROC;
IMPORT Strings;
FROM   RealConversions IMPORT RealToStr;
IMPORT TimeDate;
FROM   DOSFILEMGR  IMPORT DosOpen,  DosClose, DosWrite,  STDOUT, HFILE,
                          DosResetBuffer,
                          OPEN_ACTION_FAIL_IF_NEW,
                          OPEN_ACTION_OPEN_IF_EXISTS,
                          OPEN_SHARE_DENYNONE, FILE_NORMAL,
                          DosQueryPathInfo, DosQueryFSInfo, FSALLOCATE,
                          OPEN_ACCESS_WRITEONLY;
FROM   DOSERRORS   IMPORT ERROR_BROKEN_PIPE, NO_ERROR,
                          ERROR_PIPE_BUSY, ERROR_PIPE_NOT_CONNECTED;

CONST
   NUMSTAT = 32;
   Zero8 : EightByte = [0, 0, 0, 0, 0, 0, 0, 0];
   SB2_Pipe:  ARRAY OF CHAR = "\PIPE\DIALDMON.SB2";

TYPE
    linkstat = ( Nostat, Connect, Startlink, Up, Retry,
                 Stoplink, Disconnect, Close, Down );
    Nstat  = [0..NUMSTAT-1];
    Loadinfo = RECORD
       lstat : linkstat;
       trancnt,
       recvcnt : LONGCARD;
       END;

    SB2Prec = RECORD
       Pipe_Open : BOOLEAN;
       pname     : PSTRING;  (* name of pipe *)
       fh        : HFILE;    (* sb2 pipe handle *)
       END;

VAR
   Quiesce,
   EndConnect : BOOLEAN;
   SB2p       : ARRAY[0..0] OF SB2Prec;
   cstat      : linkstat;  (* value of last STATE *)
   rover      : Nstat;  (* current position in ldat *)
   ldat       : ARRAY Nstat OF Loadinfo;

   (* StrToNumL: convert numeric chars to long card returning
      pos of next char string or NIL and BOOLEAN for numerically ok *)
   PROCEDURE StrToNumL( src: PSTRING;
                        VAR next: PSTRING; VAR num: LONGCARD ): BOOLEAN;
    VAR
    startc,
    ch: LONGCARD;
   BEGIN
     ch := 0; num := 0;
     WHILE ( ch <= HIGH(src^) ) AND ( src^[ch] <> 0C ) AND
           ( ORD(src^[ch]) <= ORD(' ') ) DO
      INC(ch)
      END;
     IF ( ch > HIGH(src^) ) OR ( src^[ch] = 0C ) THEN RETURN FALSE END;
     IF src^[ch] = '+' THEN INC(ch) END;
     startc := ch;
     WHILE ( ch <= HIGH(src^) ) AND ( ORD(src^[ch]) >= ORD('0') ) AND
           ( ORD(src^[ch]) <= ORD('9') ) DO
           num := (10 * num) + ORD(src^[ch]) - ORD('0');
           INC(ch)
           END;
     IF ch = startc THEN RETURN FALSE END;
     IF ( ch > HIGH(src^) ) OR ( src^[ch] = 0C ) THEN
        next := NIL
     ELSE
        next := PSTRING( ADR( src^[ch] ) )
        END;

     RETURN TRUE

   END StrToNumL;

   PROCEDURE UserInt( sigcode: SHORTCARD; VAR carryon : BOOLEAN );

   BEGIN
     carryon := TRUE;   (* in all cases *)
     Quiesce := TRUE; RETURN
   END UserInt;

   (* log a time stamp *)
   PROCEDURE PutStamp;

   VAR
     cdate: CARDINAL;
     ctime: LONGCARD;
     dstr, tstr: ARRAY[0..12] OF CHAR;
     logstr: ARRAY[0..25] OF CHAR;

   BEGIN
     cdate := TimeDate.DateNow();
     ctime := TimeDate.TimeNow();
     TimeDate.DateToStr( cdate, TimeDate.Sorting, dstr );
     TimeDate.TimeToStr( ctime, tstr );
     Strings.Concat( dstr, " ", dstr );
     Strings.Concat( dstr, tstr, logstr );
     Strings.Concat( logstr, " ", logstr );
     WriteString( logstr )
   END PutStamp;

   (* put a string to the output pipe *)
   PROCEDURE Put_Pipe( pn: SHORTCARD; ostr: ARRAY OF CHAR );

     VAR
     retc:   APIRET;
     fres:   LONGCARD;

   BEGIN
          IF NOT SB2p[pn].Pipe_Open THEN
             retc := DosOpen( SB2p[pn].pname^, SB2p[pn].fh, fres, 0, FILE_NORMAL,
                     OPEN_ACTION_FAIL_IF_NEW OR
                     OPEN_ACTION_OPEN_IF_EXISTS,
                     OPEN_SHARE_DENYNONE OR
                     OPEN_ACCESS_WRITEONLY, NIL );
             IF retc = 0 THEN
                SB2p[pn].Pipe_Open := TRUE
             ELSIF retc <> ERROR_BROKEN_PIPE THEN
                 WriteLn;
                 PutStamp;
                 WriteString( "Open Error: ");
                 WrNum( retc );
                 WriteString( "  on pipe " );
                 WriteString( SB2p[pn].pname^ );
                 WriteLn;
                 DeRegExcep;
                 AbEnd( 019H )
             ELSE
                WriteLn;
                PutStamp;
                WriteString( "Broken pipe: " ); WriteString( SB2p[pn].pname^ );
                WriteLn
                END;
             END;

          IF SB2p[pn].Pipe_Open THEN
             retc := DosWrite( SB2p[pn].fh, ostr, Strings.Length(ostr), fres );
             CASE retc OF
               ERROR_BROKEN_PIPE,
               ERROR_PIPE_NOT_CONNECTED : retc := DosClose( SB2p[pn].fh );
                      WriteLn;
                      PutStamp;
                      WriteString( "Pipe broken: " );
                      WrNum( retc ); WriteLn;
                      SB2p[pn].Pipe_Open := FALSE |
               NO_ERROR: retc := DosResetBuffer( SB2p[pn].fh ) | (* sync up *)
               ELSE
                  WriteLn;
                  PutStamp;
                  WriteString( "Write Error: " );
                  WrNum( retc );
                  WriteString( "  on pipe " );
                  WriteString( SB2p[pn].pname^ ); WriteLn;
                  DeRegExcep;
                  AbEnd( 1 )
               END;
             END;

   END Put_Pipe;

CONST
  Aok : ARRAY OF CHAR = "AOK";
  Eoj : ARRAY OF CHAR = "EOJ";
  Nalit : ARRAY OF CHAR = "0.0";

VAR
  s    : Socket;
  port : CARDINAL;
  ps   : PSTRING;
  he   : hostentP;
  numstatr,
  tranr,
  recvr  : REAL;
  retc : APIRET;
  loadcnt,
  rlen,
  namelen :   LONGINT;
  tranlvl,
  recvlvl
          :   LONGCARD;
  server  :   SockAddr;
  estae   :   EXCEPTIONREGISTRATIONRECORD;
  fnumstr :   ARRAY [0..9] OF CHAR;
  loadstr :   STRING;
  buf     :   ARRAY[0..511] OF CHAR;

BEGIN (* DialdMc *)

  OutHandle := STDOUT;
  SB2p[0].Pipe_Open := FALSE;
  SB2p[0].pname  := ADR(SB2_Pipe);
  FOR rover := 0 TO NUMSTAT - 1 DO
      ldat[rover].lstat := Nostat;
      ldat[rover].trancnt := 0;
      ldat[rover].recvcnt := 0;
      END;
  cstat := Nostat;  rover := NUMSTAT - 1;
  loadcnt := 0;  numstatr := FLOAT(NUMSTAT);

  IF ( ArgCount() <> 2 ) THEN
     WriteString("Usage: DialdMc port hostname ");
     WriteLn;
     ExitProgram(1)
     END;
  ps := Arg(1);
  IF NOT StrToCard( ps^, port ) THEN
     WriteString( ps^ );
     WriteString(" port");
     ExitProgram(2)
     END;

  IF sock_init() <> 0 THEN
     psock_errno("sock_init()");
     ExitProgram(3)
     END;

  ps := Arg(2);
  he := gethostbyname( ps^ );
  IF he = NIL THEN
     WriteString( "gethostbyname failed." ); WriteLn;
     ExitProgram(12)
     END;

  s := socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
  IF s < 0 THEN
     psock_errno("socket()");
     ExitProgram(4)
     END;

  server.sin_family := AF_INET;
  server.sin_port := bswap( port );
  server.sin_addr := he^.h_addr_list^[0]^;
  server.sin_zero := Zero8;

  IF connect( s, server, SIZE(server) ) < 0 THEN
     psock_errno("connect()");
     ExitProgram(5)
     END;


  RegExcep( estae );
  RegBreak( UserInt );
  EndConnect := FALSE; Quiesce := FALSE;

  REPEAT

      rlen := recv( s, buf, SIZE(buf), 0 );
      IF rlen  < 0   THEN
         IF NOT Quiesce THEN
            psock_errno("recv()");
            EndConnect := TRUE
            (* ignore this and let the LOAD process Quiesce *)
            END
      ELSIF rlen = 0 THEN
         EndConnect := TRUE
      ELSE
         buf[rlen] := 0C;
         (* only STATE and LOAD are sent to pipe *)
         IF buf[0] = 'S' THEN
            ps := PSTRING( LONGCARD(ADR(buf[0]) + 6 ) );
            IF Strings.Compare( ps^, "CONNECT" ) = 0 THEN
               cstat := Connect
            ELSIF Strings.Compare( ps^, "START_LINK" ) = 0 THEN
               cstat := Startlink
            ELSIF Strings.Compare( ps^, "UP" ) = 0  THEN
               cstat := Up
            ELSIF Strings.Compare( ps^, "DISCONNECT" ) = 0 THEN
               cstat := Disconnect
            ELSIF Strings.Compare( ps^, "CLOSE" ) = 0 THEN
               cstat := Close
            ELSIF Strings.Compare( ps^, "RETRY" ) = 0 THEN
               cstat := Retry
            ELSIF Strings.Compare( ps^, "DOWN" )  = 0 THEN
               cstat := Down
               END;
            Put_Pipe( 0, ps^ );
            PutStamp;
            WriteString(buf); WriteLn
         ELSIF buf[0] = 'L' THEN (* respond to LOAD as heartbeat *)
            IF Quiesce THEN
               rlen := send( s, Eoj, 3, 0 ); EndConnect := TRUE
            ELSE
               INC(loadcnt);
               (* ack this xmit *)
               rlen := send( s, Aok, 3, 0 );

               (* get next rover index *)
               rover := ( rover + 1 ) MOD NUMSTAT;
               ldat[rover].lstat := cstat;
               ps := PSTRING( LONGCARD(ADR(buf[5])) );
               IF NOT StrToNumL( ps, ps, ldat[rover].trancnt ) THEN
                  ldat[rover].trancnt := 0 END;
               IF NOT StrToNumL( ps, ps, ldat[rover].recvcnt ) THEN
                  ldat[rover].recvcnt := 0 END;
               IF ( cstat = Up ) AND ( (loadcnt MOD (NUMSTAT DIV 4)) = 0 ) THEN
                  (* average load over last NUMSTAT periods *)
                  tranlvl := 0;  recvlvl := 0;
                  FOR rlen := 0 TO NUMSTAT-1 DO
                     INC( tranlvl, ldat[rlen].trancnt );
                     INC( recvlvl, ldat[rlen].recvcnt );
                     END;
                  tranr := FLOAT(tranlvl) / numstatr;
                  recvr := FLOAT(recvlvl) / numstatr;
                  loadstr[0] := 'T';  loadstr[1] := 0C;
                  IF NOT RealToStr( tranr, 2, fnumstr ) THEN
                     Strings.Assign( Nalit, fnumstr ) END;
                  Strings.Append( fnumstr, loadstr );
                  PutStamp;  WriteString( fnumstr );
                  Strings.Append( " R", loadstr );
                  IF NOT RealToStr( recvr, 2, fnumstr ) THEN
                     Strings.Assign( Nalit, fnumstr ) END;
                  WriteCh(' '); WriteString( fnumstr ); WriteLn;
                  Strings.Append( fnumstr, loadstr );
                  Put_Pipe( 0, loadstr );
                  END;
               END
         ELSE   (* ignore others *)
            END;
         END;

  UNTIL EndConnect;

  namelen := soclose(s);
  IF SB2p[0].Pipe_Open THEN
     retc := DosClose( SB2p[0].fh ) END;

  DeRegExcep;
  WriteString("DialdM Client ended successfully."); WriteLn;


END DialdMc.
