/***********************************************************/
/*                                                         */
/*  Diald Monitor  server  -- read monitor pipe            */
/*    and provide synopsis of link state over TCP port     */
/*                                                         */
/***********************************************************/
/*
 * Include Files.
 */
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <asm/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>

/*
 * Server Main.
 */
main(int argc, char **argv)
{
    unsigned short port;       /* port client will connect to              */
    struct sockaddr_in server; /* server address                           */
    struct sockaddr_in client; /* client info                              */
    int s;                     /* socket for accepting connects            */
    int ns;                    /* socket for client                        */
    int havuser, aok;          /* booleans */
    int numb,                  /* util int */
        namelen,              /* for accept */
        newlink,
        getrc,
        errlen,
        numerr,
        ctlpipe,               /* Diald control pipe */
        pipefd;                /* monitor file fd */
    FILE * pipef;              /* monitor pipe file  */
    enum { inits, down, connecting, starting, up, closing } linkst;
    char * verb;
    char * fsms;
    char moncmd[80];           /* monitor cmd */
    char buf[512];             /* data buffer for sending and receiving    */

    /*
     * Check Arguments Passed.
      Should be port, ctlpipe, monfile-pipe
     */
    if (argc != 4)
       {
        fprintf(stderr,
         "Usage: %s port diald-ctlpipe monpipe\n", argv[0]);
        exit(1);
       }


    /*
     * The host name is the first argument. Get the server address.
     */

    /*
     * The port is the second argument.
     */
    port = (unsigned short) atoi(argv[1]);

    /*
     * Put the server information into the server structure.
     * The port must be put into network byte order.
     */
    server.sin_family      = AF_INET;
    server.sin_port        = htons(port);
    server.sin_addr.s_addr = INADDR_ANY;

    /*
     * Get a stream socket.
     */
    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
       {
        perror("Socket()");
        exit(3);
       }

    /*
     * Bind the socket to the server
     */
    if ( bind( s, (struct sockaddr *)&server, sizeof(server) ) < 0 )
       {
        perror("Bind()");
        exit(4);
       }
    /* make it non-blocking 
    numb = fcntl( s, F_GETFL );
    if ( numb == -1 )
       { fprintf( stderr, "unable to get socket flags.\n" );
         exit(9);
       }
    numb &= ~O_NONBLOCK;
    numb = fcntl( s, F_SETFL, numb );
    if ( numb == -1 )
       { fprintf( stderr, "unable to set socket flags.\n" );
         exit(9);
       }  */
    printf("Socket bound.\n");

    /*
    *  Listen for connections
    */
    if ( listen( s, 1 ) != 0 )
       { perror( "Listen()" );
         exit(5);
       }

    /*
     * Open the monitor pipe-file for input
     */
    if ( ( pipefd = open( argv[3], O_RDONLY | O_NONBLOCK ) ) == -1 )
       { fprintf(stderr, "Unable to open %s err: %d\n", argv[3], errno);
         exit(3);
       }
    pipef = fdopen( pipefd, "r" );
    if ( pipef == NULL )
       { fprintf( stderr, "fdopen failure pipef.\n" );
         exit(14);
       }
    /* make the monitor pipe block */
    numb = fcntl( pipefd, F_GETFL );
    if ( numb == -1 )
       { fprintf( stderr, "unable to get pipe flags.\n" );
         exit(9);
       }
    numb &= ~O_NONBLOCK;
    numb = fcntl( pipefd, F_SETFL, numb );
    if ( numb == -1 )
       { fprintf( stderr, "unable to set pipe flags.\n" );
         exit(9);
       }
    printf("Monitor Pipe Open.\n");


    /*
    *  write the 'monitor' command to the control pipe
    */
    if ( ( ctlpipe = open( argv[2], O_WRONLY ) ) == -1 )
       { fprintf(stderr, "Unable to open %s err: %d\n", argv[2], errno);
         exit(7);
       }
    strcpy( moncmd, "monitor "); strcat( moncmd, argv[3] );
    strcat( moncmd, "\n" );
    numb = write( ctlpipe, moncmd, strlen(moncmd) );
    if ( numb != (strlen(moncmd)))
       { fprintf(stderr, "monitor command not written.\n");
         exit(8);
       }
    close( ctlpipe );
    printf("Monitor command sent to diald.\n");
    

    aok = 1;  havuser = 0; linkst = inits;
    while ( aok )
      {
      if (!havuser)
         { namelen = sizeof(client);
           ns = accept( s, (struct sockaddr *)&client, &namelen );
           if ( ns == -1 )
              { errlen = sizeof(numerr);
                numb = getsockopt( s, SOL_SOCKET, SO_ERROR,
                  &numerr, &errlen );
                if ( numb == -1 ) 
                   { fprintf( stderr, "SO_ERROR get failed.\n" );
                     exit(10);
                   }
                if ( numerr != EWOULDBLOCK )
                   { perror("accept()");
                     exit(11);
                   }
              }
           else
              {
              fprintf( stderr, "Successful accept\n" );
              havuser = 1;  newlink = 1;
              }
         } /* ifnot  havuser */
      verb = fgets( buf, sizeof(buf), pipef );
      if ( verb == NULL )
         { aok = 0;
           fprintf( stderr, "ferror(): %d\n" );
         }
      else
         {
         namelen = strlen(buf) - 1;
         buf[namelen] = '\0';  /* strip lf */
         if ( havuser )
            {
            /* analyze monitor line for xmit info */
            if ( verb != NULL )
               { 
               numb = 0;
               if ( strcmp( verb, "STATUS" ) == 0 )
                  { 
                  }
               else
               if ( strcmp( verb, "LOAD" ) == 0 )
                  { 
                  if ( newlink )
                     { newlink = 0;
                       strcpy( moncmd, "STATE " );
                       if ( linkst == up ) strcat( moncmd, "UP" );
                       else strcat( moncmd, "DOWN" );
                       numerr =  send( ns, moncmd, strlen(moncmd)+1, 0 );
                     }
                  strcpy( moncmd, verb );
                  moncmd[4] = ' '; moncmd[5] = '\0';
                  verb = fgets( buf, sizeof(buf), pipef );
                  if ( verb != NULL )
                     {
                     numb = strlen(buf) - 1;
                     buf[numb] = ' '; /* chng lf */
                     strcat( moncmd, buf );
                     verb = fgets( buf, sizeof(buf), pipef );
                     if ( verb != NULL )
                        { 
                          numb = strlen(buf) - 1;
                          buf[numb] = ' '; /* chng lf */
                          strcat( moncmd, buf );
                        }
                     numb = strlen(moncmd);
                     }
                  }
               else
               if ( strcmp( verb, "QUEUE" ) == 0 )
                  {
                  }
               else
               if ( strcmp( verb, "STATE" ) == 0 )
                  {
                  newlink = 0;
                  strcpy( moncmd, verb );
                  moncmd[5] = ' '; moncmd[6] = '\0';
                  verb = fgets( buf, sizeof(buf), pipef );
                  if ( verb != NULL )
                     {
                     numb = strlen(buf) - 1;
                     buf[numb] = '\0'; /* strip lf */
                     if ( strcmp( buf, "UP" ) == 0 )
                        {
                        linkst = up;
                        }
                     else 
                        linkst = down;
                     strcat( moncmd, buf );
                     numb = strlen(moncmd);
                     }
                  else
                     { aok = 0;
                       fprintf( stderr, "ferror(): %d\n" );
                       numb = 0;
                     }
                  }
               else numb = 0;  /* ignore others */
               }
            if ( numb > 0 )
               {
               numerr =  send( ns, moncmd, numb, 0 );
               if ( numerr < 0 )
                  {
                  perror("Send()");
                  exit(5);
                  }
               errlen = sizeof(numerr);
               getrc = getsockopt( ns, SOL_SOCKET, SO_ERROR,
                  &numerr, &errlen );
               if ( getrc == -1 )
                   { fprintf( stderr, "SO_ERROR get failed.\n" );
                     exit(10);
                   }
               if ( numerr != 0 )
                   { fprintf( stderr, "numerr: %d\n", numerr );
                     perror("pre-send()");
                     havuser = 0;
                   }
               if ( moncmd[0] == 'L' )
                  { numerr = recv( ns, buf, 3, 0 );
                    if ( buf[0] != 'A' )
                       { printf( "User disconnecting.\n");
                       havuser = 0;
                       }
                  }
               }
            } /* havuser */
         } /* is aok */
      } /* end while */

    /*
     * Close the socket and monitor.
     */
    printf("Closing Monitor Pipe.\n");
    close( pipefd );
    close(s);

    printf("DialdM Server Ended Successfully\n");
    exit(0);
}
